"""
Common code for the pyslim test cases.
"""
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import division

import random
import unittest
import base64


def setUp():
    # Make random tests reproducible.
    random.seed(210)


class PyslimTestCase(unittest.TestCase):
    '''
    Base class for test cases in pyslim.
    '''

    def assertArrayEqual(self, x, y):
        self.assertListEqual(list(x), list(y))

    def assertArrayAlmostEqual(self, x, y):
        self.assertEqual(len(x), len(y))
        for a, b in zip(x, y):
            self.assertAlmostEqual(a, b)

    def assertTablesAlmostEqual(self, t1, t2):
        # nodes
        self.assertEqual(t1.nodes.num_rows, t2.nodes.num_rows)
        self.assertArrayEqual(t1.nodes.flags, t2.nodes.flags)
        self.assertArrayEqual(t1.nodes.individual, t2.nodes.individual)
        self.assertArrayEqual(t1.nodes.metadata, t2.nodes.metadata)
        self.assertArrayEqual(t1.nodes.metadata_offset, t2.nodes.metadata_offset)
        self.assertArrayEqual(t1.nodes.population, t2.nodes.population)
        self.assertArrayAlmostEqual(t1.nodes.time, t2.nodes.time)
        # other tables
        self.assertEqual(t1.edges, t2.edges)
        self.assertEqual(t1.sites, t2.sites)
        self.assertEqual(t1.mutations, t2.mutations)
        self.assertEqual(t1.migrations, t2.migrations)
        self.assertEqual(t1.individuals, t2.individuals)
        self.assertEqual(t1.provenances, t2.provenances)

    def verify_haplotype_equality(self, ts, slim_ts):
        self.assertEqual(ts.num_sites, slim_ts.num_sites)
        for j, v1, v2 in zip(range(ts.num_sites), ts.variants(),
                             slim_ts.variants()):
            g1 = [v1.alleles[x] for x in v1.genotypes]
            g2 = [v2.alleles[x] for x in v2.genotypes]
            self.assertArrayEqual(g1, g2)



import attr
import struct
import msprime
import json
from collections import OrderedDict

from .slim_metadata import *
from .provenance import *

SLIM_VERSION = "3.0"
SLIM_FILE_VERSION = "0.1"

def load(path, slim_format):
    '''
    Load the tree sequence found in the .trees file at ``path``. If the .trees
    file is SLiM-compatible, set ``slim_format`` to ``True`` (in which case
    this returns a :class:`SlimTreeSequence`); otherwise, this just calls
    :meth:`msprime.load`.  A SlimTreeSequence has all node and migration times
    in the tree sequence shifted relative to that recorded in the file by the
    current generation recorded by SLiM so that the tskit times are measured in
    units of generations before the end of the simulation.

    :param string path: The path to a .trees file.
    :param bool slim_format: Whether the .trees file should be coverted from
        SLiM format.
    '''
    if slim_format:
        ts = SlimTreeSequence.load(path)
    else:
        ts = msprime.load(path)
    return ts


def load_tables(tables, slim_format):
    '''
    See :func:`load`.

    :param TableCollection tables: A set of tables.
    :param bool slim_format: Whether the tables should be coverted from
        SLiM format.
    '''
    if slim_format:
        ts = SlimTreeSequence.load_tables(tables)
    else:
        ts = tables.tree_sequence()
    return ts


def annotate_defaults(ts, model_type, slim_generation, remembered_node_count=0):
    '''
    Takes a tree sequence (as produced by msprime, for instance), and adds in the
    information necessary for SLiM to use it as an initial state, filling in
    mostly default values. Returns a :class:`SlimTreeSequence`.

    :param TreeSequence ts: A :class:`TreeSequence`.
    :param string model_type: SLiM model type: either "WF" or "nonWF".
    :param int slim_generation: What generation number in SLiM correponds to
        ``time=0`` in the tree sequence.
    :param int remembered_node_count: (NOT SUPPORTED) How many nodes in the
        tree sequence will be marked as "ancestral samples" by SLiM (it must by the
        *first* this many nodes).  
    '''
    tables = ts.tables
    annotate_defaults_tables(tables, model_type, slim_generation, remembered_node_count)
    return SlimTreeSequence.load_tables(tables)


def annotate_defaults_tables(tables, model_type, slim_generation, remembered_node_count=0):
    '''
    Does the work of :func:`annotate_defaults()`, but modifies the tables in place: so,
    takes tables as produced by ``msprime``, and makes them look like the
    tables as output by SLiM. See :func:`annotate_defaults` for details.
    '''
    if remembered_node_count != 0:
        raise ValueError("Setting remembered node count not currently supported.")
    if (type(slim_generation) is not int) or (slim_generation < 1):
        raise ValueError("SLiM generation must be an integer and at least 1.")
    # set_nodes must come before set_populations
    if model_type == "WF":
        default_ages = -1
    elif model_type == "nonWF":
        default_ages = 0
    else:
        raise ValueError("Model type must be 'WF' or 'nonWF'")
    _set_nodes_individuals(tables, age=default_ages)
    _set_populations(tables)
    _set_sites_mutations(tables)
    _set_provenance(tables, model_type=model_type, slim_generation=slim_generation,
                   remembered_node_count=remembered_node_count)


class SlimTreeSequence(msprime.TreeSequence):
    '''
    This is just like a :class:`TreeSequence`, except that:
        - Times are shifted by the `generation` in the last SLiM entry
            of the Provenance table.
    You should create a :class:`SlimTreeSequence` using one of
    :meth:`SlimTreeSequence.load_tables` :meth:`SlimTreeSequence.load`,
    :func:`load`, or :func:`load_tables`.

    :ivar slim_generation: The amount by which times have been shifted.
    :vartype slim_generation: int
    '''

    def __init__(self, ts, slim_generation):
        self._ll_tree_sequence = ts._ll_tree_sequence
        self.slim_generation = slim_generation


    @classmethod
    def load(cls, path):
        '''
        Load a :class:`SlimTreeSequence` from a .trees file on disk.

        :param string path: The path to a .trees file.
        :rtype SlimTreeSequence:
        '''
        # roundabout way to load just the tables
        ts = msprime.load(path)
        tables = ts.dump_tables()
        provenance = get_provenance(tables)
        _set_slim_generation(tables, provenance.slim_generation)
        return cls.load_tables(tables)

    @classmethod
    def load_tables(cls, tables):
        '''
        Creates the :class:`SlimTreeSequence` defined by the tables.

        :param TableCollection tables: A set of tables, as produced by SLiM
            or by annotate_defaults().
        :rtype SlimTreeSequence:
        '''
        # a roundabout way to copy the tables
        ts = tables.tree_sequence()
        new_tables = ts.dump_tables()
        provenance = get_provenance(new_tables)
        ts = msprime.TableCollection.tree_sequence(new_tables)
        return cls(ts, provenance.slim_generation)

    def dump(self, path, **kwargs):
        '''
        Write out the .trees file that can be read back in by SLiM. See
        :meth:`msprime.TreeSequence.dump()` for other arguments.

        :param string path: The path to a .trees file.
        '''
        # This would be simpler if there were a python-level TableCollection.dump
        # method: https://github.com/tskit-dev/msprime/issues/547
        tables = self.dump_tables()
        _set_slim_generation(tables, -1 * self.slim_generation)
        temp_ts = msprime.TableCollection.tree_sequence(tables)
        msprime.TreeSequence.dump(temp_ts, path, **kwargs)


def _set_slim_generation(tables, slim_generation):
    '''
    Modifying ``tables`` in place, shifts the "time ago" entries in the tables
    to be measured in units of time *before* `slim_generation`, by adding
    ``slim_generation`` to the ``time`` columns of Node and Migration tables.
    Can be inverted by passing in ``-1 * slim_generation``.
    '''
    tables.nodes.set_columns(flags=tables.nodes.flags,
            time=tables.nodes.time + slim_generation,
            population=tables.nodes.population, individual=tables.nodes.individual,
            metadata=tables.nodes.metadata, metadata_offset=tables.nodes.metadata_offset)
    tables.migrations.set_columns(left=tables.migrations.left, right=tables.migrations.right,
            node=tables.migrations.node, source=tables.migrations.source,
            dest=tables.migrations.dest, time=tables.migrations.time + slim_generation)


def _set_nodes_individuals(
        tables, node_ind=None, location=(0, 0, 0), age=0, ind_id=None,
        ind_population=None, ind_sex=INDIVIDUAL_TYPE_HERMAPHRODITE,
        ind_flags=0, slim_ind_flags=0, node_id=None,
        node_is_null=False, node_type=GENOME_TYPE_AUTOSOME):
    '''
    Adds to a TableCollection the information relevant to individuals required
    for SLiM to load in a tree sequence, that is found in Node and Individual
    tables.  This will replace any existing Individual table, and will replace
    any information already in the individual, metadata, and population columns
    of the Node table.

    This is designed to make it easy to assign default values:
    - (node_ind) the 2*j-th and (2*j+1)-st `sample` nodes to individual j
    - (location) individual locations to (0, 0, 0)
    - (age) individual age to 0
    - (ind_id) SLiM individual pedigree IDs to sequential integers starting from 0
    - (ind_population) individual populations to 0
    - (node_id) SLiM genome IDs to sequential integers starting with samples from 0
    - (node_is_null) genomes to be non-null
    - (node_type) genome type to 0 (= autosome)
    '''
    samples = list(filter(lambda j: tables.nodes.flags[j] & msprime.NODE_IS_SAMPLE,
                          range(tables.nodes.num_rows)))
    if (len(samples) % 2) != 0:
        raise ValueError("There must be an even number of sampled nodes,"\
                         + "since organisms are diploid.")

    if node_ind is None:
        node_ind = [msprime.NULL_INDIVIDUAL for _ in range(tables.nodes.num_rows)]
        for j, k in enumerate(samples):
            node_ind[j] = int(k/2)

    num_individuals = max(node_ind) + 1
    num_nodes = tables.nodes.num_rows

    if type(location) is tuple:
        location = [location for _ in range(num_individuals)]
    assert(len(location) == num_individuals)

    if type(age) is int or type(age) is float:
        age = [age for _ in range(num_individuals)]
    assert(len(age) == num_individuals)

    if ind_id is None:
        ind_id = list(range(num_individuals))
    assert(len(ind_id) == num_individuals)

    if type(ind_sex) is int:
        ind_sex = [ind_sex for _ in range(num_individuals)]
    assert(len(ind_sex) == num_individuals)

    if type(slim_ind_flags) is int:
        slim_ind_flags = [slim_ind_flags for _ in range(num_individuals)]
    assert(len(slim_ind_flags) == num_individuals)

    if type(ind_flags) is int:
        ind_flags = [ind_flags for _ in range(num_individuals)]
    assert(len(ind_flags) == num_individuals)

    if node_id is None:
        node_id = [-1 for _ in range(num_nodes)]
        for j, k in enumerate(list(samples)
                              + sorted(list(set(range(num_nodes))
                                            - set(samples)))):
            node_id[k] = j
    assert(len(node_id) == num_nodes)

    if type(node_is_null) is bool:
        node_is_null = [node_is_null for _ in range(num_nodes)]
    assert(len(node_is_null) == num_nodes)

    if type(node_type) is int:
        node_type = [node_type for _ in range(num_nodes)]
    assert(len(node_type) == tables.nodes.num_rows)

    if ind_population is None:
        # set the individual populations based on what's in the nodes
        ind_population = [msprime.NULL_POPULATION for _ in range(num_individuals)]
        for j, u in enumerate(node_ind):
            ind_population[u] = tables.nodes.population[j]
    assert(len(ind_population) == num_individuals)

    # check for consistency: every individual has two nodes, and populations agree
    ploidy = [0 for _ in range(num_individuals)]
    for j in samples:
        u = node_ind[j]
        assert(u >= 0)
        ploidy[u] += 1
        if tables.nodes.population[j] != ind_population[u]:
            raise ValueError("Inconsistent populations: nodes and individuals do not agree.")

    if any([p != 2 for p in ploidy]):
        raise ValueError("Not all individuals have two assigned nodes.")

    tables.nodes.set_columns(flags=tables.nodes.flags, time=tables.nodes.time,
                             population=tables.nodes.population, individual=node_ind,
                             metadata=tables.nodes.metadata,
                             metadata_offset=tables.nodes.metadata_offset)

    loc_vec, loc_off = msprime.pack_bytes(location)
    tables.individuals.set_columns(
            flags=ind_flags, location=loc_vec, location_offset=loc_off)

    individual_metadata = [IndividualMetadata(*x) for x in
                           zip(ind_id, age, ind_population, ind_sex, slim_ind_flags)]
    node_metadata = [None for _ in range(num_nodes)]
    for j in samples:
        node_metadata[j] = NodeMetadata(slim_id=node_id[j], is_null=node_is_null[j],
                                        genome_type=node_type[j])

    annotate_individual_metadata(tables, individual_metadata)
    annotate_node_metadata(tables, node_metadata)


def _set_populations(
        tables, pop_id=None, selfing_fraction=0.0, female_cloning_fraction=0.0,
        male_cloning_fraction=0.0, sex_ratio=0.5, bounds_x0=0.0, bounds_x1=0.0,
        bounds_y0=0.0, bounds_y1=0.0, bounds_z0=0.0, bounds_z1=0.0,
        migration_records=None):
    '''
    Adds to a TableCollection the information about populations required for SLiM
    to load a tree sequence. This will replace anything already in the Population
    table.
    '''
    num_pops = max(tables.nodes.population) + 1
    for md in msprime.unpack_bytes(tables.individuals.metadata,
                                   tables.individuals.metadata_offset):
        try:
            ind_md = decode_individual(md)
        except:
            raise ValueError("Individuals do not have metadata:"
                    + "need to run set_nodes_individuals() first?")
        assert(ind_md.population < num_pops)
    if pop_id is None:
        pop_id = list(range(num_pops))
    assert(len(pop_id) == num_pops)

    if type(selfing_fraction) is float:
        selfing_fraction = [selfing_fraction for _ in range(num_pops)]
    assert(len(selfing_fraction) == num_pops)

    if type(female_cloning_fraction) is float:
        female_cloning_fraction = [female_cloning_fraction for _ in range(num_pops)]
    assert(len(female_cloning_fraction) == num_pops)

    if type(male_cloning_fraction) is float:
        male_cloning_fraction = [male_cloning_fraction for _ in range(num_pops)]
    assert(len(male_cloning_fraction) == num_pops)

    if type(sex_ratio) is float:
        sex_ratio = [sex_ratio for _ in range(num_pops)]
    assert(len(sex_ratio) == num_pops)

    if type(bounds_x0) is float:
        bounds_x0 = [bounds_x0 for _ in range(num_pops)]
    assert(len(bounds_x0) == num_pops)

    if type(bounds_x1) is float:
        bounds_x1 = [bounds_x1 for _ in range(num_pops)]
    assert(len(bounds_x1) == num_pops)

    if type(bounds_y0) is float:
        bounds_y0 = [bounds_y0 for _ in range(num_pops)]
    assert(len(bounds_y0) == num_pops)

    if type(bounds_y1) is float:
        bounds_y1 = [bounds_y1 for _ in range(num_pops)]
    assert(len(bounds_y1) == num_pops)

    if type(bounds_z0) is float:
        bounds_z0 = [bounds_z0 for _ in range(num_pops)]
    assert(len(bounds_z0) == num_pops)

    if type(bounds_z1) is float:
        bounds_z1 = [bounds_z1 for _ in range(num_pops)]
    assert(len(bounds_z1) == num_pops)

    if migration_records is None:
        migration_records = [[] for _ in range(num_pops)]
    assert(len(migration_records) == num_pops)
    for mrl in migration_records:
        for mr in mrl:
            assert(type(mr) is PopulationMigrationMetadata)

    population_metadata = [PopulationMetadata(*x) for x in
                           zip(pop_id, selfing_fraction, female_cloning_fraction,
                               male_cloning_fraction, sex_ratio, bounds_x0,
                               bounds_x1, bounds_y0, bounds_y1, bounds_z0, bounds_z1,
                               migration_records)]
    annotate_population_metadata(tables, population_metadata)


def _set_sites_mutations(
        tables, mutation_id=None, mutation_type=1, selection_coeff=0.0,
        population=msprime.NULL_POPULATION, slim_time=None):
    '''
    Adds to a TableCollection the information relevant to mutations required
    for SLiM to load in a tree sequence. This means adding to the metadata column
    of the Mutation table,  It will also
    - give SLiM IDs to each mutation
    - round Site positions to integer values
    - stack any mutations that end up at the same position as a result
    - replace ancestral states with ""
    This will replace any information already in the metadata or derived state
    columns of the Mutation table.
    '''
    num_mutations = tables.mutations.num_rows

    if mutation_id is None:
        mutation_id = list(range(num_mutations))
    assert(len(mutation_id) == num_mutations)

    if type(mutation_type) is int:
        mutation_type = [mutation_type for _ in range(num_mutations)]
    assert(len(mutation_type) == num_mutations)

    if type(selection_coeff) is float:
        selection_coeff = [selection_coeff for _ in range(num_mutations)]
    assert(len(selection_coeff) == num_mutations)

    if type(population) is int:
        population = [population for _ in range(num_mutations)]
    assert(len(population) == num_mutations)

    if slim_time is None:
        ## This may *not* make sense because we have to round:
        # slim_time = [(-1) * int(tables.nodes.time[u]) for u in tables.mutations.node]
        slim_time = [0 for _ in range(num_mutations)]
    assert(len(slim_time) == num_mutations)

    mutation_metadata = [[MutationMetadata(*x)] for x in
                         zip(mutation_type, selection_coeff, population, slim_time)]
    annotate_mutation_metadata(tables, mutation_metadata)

#######
# Provenance
####################
# The general structure of a Provenance entry is a JSON string:
# {“program”:“SLiM”, “version”:“<version>“, “file_version”:“<file_version>“,
#     “model_type”:“<model_type>“, “generation”:<generation>,
#     “remembered_node_count”:<rem_count>}
# The field values in angle brackets have the following meanings:
# <version>: The version of SLiM that generated the file, such as 3.0.
# <file_version>: The metadata format of the file; at present only 0.1 is supported.
# <model_type>: This should be either WF or nonWF, depending upon the type of
# model that generated the file.  This has some implications for the other
# metadata; in particular, some of the population metadata is required for WF
# models but unused in nonWF models, and individual ages in WF model data are
# expected to be -1.  Note that SLiM will allow WF model data to be read into a
# nonWF model, and vice versa, but since this is usually not intentional a
# warning will be issued.  Moving data between model types has not been tested,
# so be aware that issues may exist with doing so.
# <generation>: The generation number, which will be set by SLiM upon loading.
# <rem_count>: The number of remembered nodes, in addition to the sample.


@attr.s
class ProvenanceMetadata(object):
    model_type = attr.ib()
    slim_generation = attr.ib()
    remembered_node_count = attr.ib()


def get_provenance(tables):
    '''
    Extracts model type, slim generation, and remembmered node count from the last
    entry in the provenance table that is tagged with "program"="SLiM".

    :param TableCollection tables: The tables.
    :rtype ProvenanceMetadata:
    '''
    prov = [json.loads(x.record) for x in tables.provenances]
    slim_prov = [u for u in prov if ('program' in u and u['program'] == "SLiM")]
    if len(slim_prov) == 0:
        raise ValueError("Tree sequence contains no SLiM provenance entries.")
    last_slim_prov = slim_prov[len(slim_prov)-1]
    return ProvenanceMetadata(last_slim_prov["model_type"], last_slim_prov["generation"],
                              last_slim_prov["remembered_node_count"])


def _set_provenance(tables, model_type, slim_generation, remembered_node_count=0):
    '''
    Appends to the provenance table of a :class:`TableCollection` a record containing
    the information that SLiM expects to find there.

    :param TableCollection tables: The table collection.
    :param string model_type: The model type: either "WF" or "nonWF".
    :param int slim_generation: The "current" generation in the SLiM simulation.
    :param int remembered_node_count: The number of nodes that will be "remembered".
    '''
    pyslim_provenance = get_provenance_dict()
    pyslim_dict = OrderedDict([("program", "pyslim"),
                               ("version", pyslim_provenance['version'])])
    slim_dict = OrderedDict([
                 ("program", "SLiM"), ("version", SLIM_VERSION),
                 ("file_version", SLIM_FILE_VERSION),
                 ("model_type", model_type), ("generation", slim_generation),
                 ("remembered_node_count", remembered_node_count)])
    tables.provenances.add_row(json.dumps(pyslim_dict))
    tables.provenances.add_row(json.dumps(slim_dict))




import attr
import struct
import msprime
import json

GENOME_TYPE_AUTOSOME = 0
GENOME_TYPE_X = 1
GENOME_TYPE_Y = 2
INDIVIDUAL_TYPE_HERMAPHRODITE = -1
INDIVIDUAL_TYPE_FEMALE = 0
INDIVIDUAL_TYPE_MALE = 1
INDIVIDUAL_FLAG_MIGRATED = 0x01

###########
# Mutations
#  The metadata for a Mutation is a *collection* of these structs,
#  thanks to mutation stacking.
###########
# typedef struct __attribute__((__packed__)) {
#     slim_objectid_t mutation_type_id_;    // 4 bytes (int32_t): the id of the mutation type the mutation belongs to
#     slim_selcoeff_t selection_coeff_;     // 4 bytes (float): the selection coefficient
#     slim_objectid_t subpop_index_;        // 4 bytes (int32_t): the id of the subpopulation in which the mutation arose
#     slim_generation_t origin_generation_; // 4 bytes (int32_t): the generation in which the mutation arose
# } MutationMetadataRec;
#

@attr.s
class MutationMetadata(object):
    mutation_type = attr.ib()
    selection_coeff = attr.ib()
    population = attr.ib()
    slim_time = attr.ib()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

def decode_mutation(buff):
    '''
    Extracts the information stored in binary by SLiM in the ``metadata``
    column of a :class:`MutationTable`.  If ``buff`` is empty, returns ``[]``.

    :param bytes buff: The ``metadata`` entry of a row of a
        :class:`MutationTable`, as stored by SLiM.
    :rtype list:
    '''
    # note that in the case that buff is of length zero
    # this returns [] instead of None, like the others do
    num_muts = int(len(buff) / 16) # 4 + 4 + 4 + 4
    if len(buff) != num_muts * 16:
        raise ValueError("Mutation metadata of incorrect format.")
    struct_string = "<" + "ifii" * num_muts
    metadata = struct.unpack(struct_string, buff)
    mut_structs = []
    for k in range(num_muts):
        mutation_type = metadata[k * 4]
        selection_coeff = metadata[k * 4 + 1]
        population = metadata[k * 4 + 2]
        slim_time = metadata[k * 4 + 3]
        mut_structs.append(MutationMetadata(mutation_type=mutation_type,
                                            selection_coeff=selection_coeff,
                                            population=population,
                                            slim_time=slim_time))
    return mut_structs

def encode_mutation(metadata_object):
    '''
    Encodes the list of :class:`MutationMetadata` objects as a bytes object,
    suitable to be put in as metadata for a mutation.  Unlike other ``encode_`
    functions, this takes a *list* rather than a single value.

    :param MutationMetadata metadata_object: The list of
        :class:`MutationMetadata` objects to be encoded.  
    :rtype bytes:
    '''
    mr_values = []
    for mr in metadata_object:
        mr_values.extend([mr.mutation_type, mr.selection_coeff,
                          mr.population, mr.slim_time])
    struct_string = "<" + "ifii" * len(metadata_object)
    return struct.pack(struct_string, *mr_values)


def extract_mutation_metadata(tables):
    '''
    Returns an iterator over lists of :class:`MutationMetadata` objects containing
    information about the mutations in the tables.

    :param TableCollection tables: The tables, as produced by SLiM.
    '''
    metadata = msprime.unpack_bytes(tables.mutations.metadata,
                                    tables.mutations.metadata_offset)
    for md in metadata:
        yield decode_mutation(md)


def annotate_mutation_metadata(tables, metadata):
    '''
    Revise the mutation table so that the metadata column is given by applying
    `encode_mutation()` to the sources given.

    :param TableCollection tables: a table collection to be modified
    :param iterable metadata: a list of (lists of MutationMetadata) or None objects
    '''
    if len(metadata) != tables.mutations.num_rows:
        raise ValueError("annotate mutations: metadata not the same length as the table.")
    orig_mutations = tables.mutations.copy()
    tables.mutations.clear()
    for mut, md in zip(orig_mutations, metadata):
        if md is None:
            metadata = b''
        else:
            metadata = encode_mutation(md)
        tables.mutations.add_row(site=mut.site, node=mut.node, derived_state=mut.derived_state,
                                 parent=mut.parent, metadata=metadata)

#######
# Nodes
#######
# typedef struct __attribute__((__packed__)) {
#     slim_genomeid_t genome_id_; // 8 bytes (int64_t): the SLiM genome ID for this genome, assigned by pedigree rec
#     uint8_t is_null_;           // 1 byte (uint8_t): true if this is a null genome (should never contain mutations)
#     GenomeType type_;           // 1 byte (uint8_t): the type of the genome (A, X, Y)
# } GenomeMetadataRec;
#

_node_struct = struct.Struct("<qBB")

@attr.s
class NodeMetadata(object):
    slim_id = attr.ib()
    is_null = attr.ib()
    genome_type = attr.ib()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

def decode_node(buff):
    '''
    Extracts the information stored in binary by SLiM in the ``metadata``
    column of a :class:`NodeTable`.  If the buffer is empty, returns None.

    :param bytes buff: The ``metadata`` entry of a row of a
        :class:`NodeTable`, as stored by SLiM.
    :rtype NodeMetadata:
    '''
    if len(buff) == 0:
        md = None
    else:
        if len(buff) != 10: # 8 + 1 + 1
            raise ValueError("Node metadata of incorrect format.")
        slim_id, is_null, genome_type = _node_struct.unpack(buff)
        md = NodeMetadata(slim_id=slim_id, is_null=is_null, genome_type=genome_type)
    return md

def encode_node(metadata_object):
    '''
    Encodes :class:`NodeMetadata` objects as a bytes object, suitable to be put
    in as metadata for a node. If ``metadata_object`` is ``None``, returns an
    empty bytes object.

    :param NodeMetadata metadata_object: The object to be encoded.
    :rtype bytes:
    '''
    if metadata_object is None:
        md = b''
    else:
        md = _node_struct.pack(metadata_object.slim_id, metadata_object.is_null,
                               metadata_object.genome_type)
    return md


def extract_node_metadata(tables):
    '''
    Returns an iterator over lists of :class: `NodeMetadata` objects containing
    information about the nodes in the tables.

    :param TableCollection tables: The tables, as produced by SLiM.
    '''
    metadata = msprime.unpack_bytes(tables.nodes.metadata,
                                    tables.nodes.metadata_offset)
    for md in metadata:
        yield decode_node(md)


def annotate_node_metadata(tables, metadata):
    '''
    Modify the NodeTable so that the metadata
    column is given by applying `encode_node()` to the sources given.

    :param TableCollection tables: a table collection to be modified
    :param iterable metadata: a list of NodeMetadata or None objects
    '''
    if len(metadata) != tables.nodes.num_rows:
        raise ValueError("annotate nodes: metadata not the same length as the table.")
    orig_nodes = tables.nodes.copy()
    tables.nodes.clear()
    for node, md in zip(orig_nodes, metadata):
        if md is None:
            metadata = b''
        else:
            metadata = encode_node(md)
        tables.nodes.add_row(flags=node.flags, time=node.time, population=node.population,
                             individual=node.individual, metadata=metadata)


#######
# Individuals
#######
# typedef struct __attribute__((__packed__)) {
#   slim_pedigreeid_t pedigree_id_;    // 8 bytes (int64_t): the SLiM pedigree ID for this individual, assigned by pedigree rec
#   slim_age_t age_;                   // 4 bytes (int32_t): the age of the individual (-1 for WF models)
#   slim_objectid_t subpopulation_id_; // 4 bytes (int32_t): the subpopulation the individual belongs to
#   IndividualSex sex_;    // 4 bytes (int32_t): the sex of the individual, as defined by the IndividualSex enum
#   uint32_t flags_;       // 4 bytes (uint32_t): assorted flags, see below
# } IndividualMetadataRec;

_individual_struct = struct.Struct("<qiiiI")

@attr.s
class IndividualMetadata(object):
    pedigree_id = attr.ib()
    age = attr.ib()
    population = attr.ib()
    sex = attr.ib()
    flags = attr.ib()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

def decode_individual(buff):
    '''
    Extracts the information stored in binary by SLiM in the ``metadata``
    column of a :class:`IndividualTable`. If the buffer is empty, returns None.

    :param bytes buff: The ``metadata`` entry of a row of a
        :class:`IndividualTable`, as stored by SLiM.
    :rtype IndividualMetadata:
    '''
    if len(buff) == 0:
        md = None
    else:
        if len(buff) != 24: # 8 + 4 + 4 + 4 + 4:
            raise ValueError("Individual metadata of incorrect format.")
        pedigree_id, age, population, sex, flags = _individual_struct.unpack(buff)
        md = IndividualMetadata(
                    pedigree_id=pedigree_id, age=age, population=population,
                    sex=sex, flags=flags)
    return md

def encode_individual(metadata_object):
    '''
    Encodes :class:`IndividualMetadata` objects as a bytes object, suitable to be put
    in as metadata for an individual.  If ``buff`` is ``None``, returns an
    empty bytes object. 

    :param IndividualMetadata metadata_object: The object to be encoded.
    :rtype bytes:
    '''
    if metadata_object is None:
        md = b''
    else:
        md = _individual_struct.pack(metadata_object.pedigree_id, metadata_object.age,
                                     metadata_object.population, metadata_object.sex,
                                     metadata_object.flags)
    return md


def extract_individual_metadata(tables):
    '''
    Returns an iterator over lists of :class:`IndividualMetadata` objects
    containing information about the individuals in the tables.

    :param TableCollection tables: The tables, as produced by SLiM.
    '''
    metadata = msprime.unpack_bytes(tables.individuals.metadata,
                                    tables.individuals.metadata_offset)
    for md in metadata:
        yield decode_individual(md)


def annotate_individual_metadata(tables, metadata):
    '''
    Modify the IndividualTable so that the metadata
    column is given by applying `encode_individual()` to the sources given.

    :param TableCollection tables: a table collection to be modified
    :param iterable metadata: a list of IndividualMetadata or None objects
    '''
    if len(metadata) != tables.individuals.num_rows:
        raise ValueError("annotate individuals: metadata not the same length as the table.")
    orig_individuals = tables.individuals.copy()
    tables.individuals.clear()
    for ind, md in zip(orig_individuals, metadata):
        if md is None:
            metadata = b''
        else:
            metadata = encode_individual(md)
        tables.individuals.add_row(flags=ind.flags, location=ind.location, metadata=metadata)

#######
# Populations
####################
# typedef struct __attribute__((__packed__)) {
#     slim_objectid_t subpopulation_id_; // 4 bytes (int32_t): the id of this subpopulation
#     double selfing_fraction_;          // 8 bytes (double): selfing fraction (unused in non-sexual models), unused in nonWF models
#     double female_clone_fraction_;     // 8 bytes (double): cloning fraction (females / hermaphrodites), unused in nonWF models
#     double male_clone_fraction_;       // 8 bytes (double): cloning fraction (males / hermaphrodites), unused in nonWF models
#     double sex_ratio_;                 // 8 bytes (double): sex ratio (M:M+F), unused in nonWF models
#     double bounds_x0_;                 // 8 bytes (double): spatial bounds, unused in non-spatial models
#     double bounds_x1_;                 // 8 bytes (double): spatial bounds, unused in non-spatial models
#     double bounds_y0_;                 // 8 bytes (double): spatial bounds, unused in non-spatial / 1D models
#     double bounds_y1_;                 // 8 bytes (double): spatial bounds, unused in non-spatial / 1D models
#     double bounds_z0_;                 // 8 bytes (double): spatial bounds, unused in non-spatial / 1D / 2D models
#     double bounds_z1_;                 // 8 bytes (double): spatial bounds, unused in non-spatial / 1D / 2D models
#     int32_t migration_rec_count_;      // 4 bytes (int32_t): the number of migration records, 0 in nonWF models
#     // followed by migration_rec_count_ instances of SubpopulationMigrationMetadataRec
# } SubpopulationMetadataRec;
#
# typedef struct __attribute__((__packed__)) {
#     slim_objectid_t source_subpop_id_; // 4 bytes (int32_t): the id of the source subpopulation, unused in nonWF models
#     double migration_rate_;            // 8 bytes (double): the migration rate from source_subpop_id_, unused in nonWF models
# } SubpopulationMigrationMetadataRec;

@attr.s
class PopulationMigrationMetadata(object):
    source_subpop = attr.ib()
    migration_rate = attr.ib()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    def __iter__(self):
        for a in (self.source_subpop, self.migration_rate):
            yield a

@attr.s
class PopulationMetadata(object):
    slim_id = attr.ib()
    selfing_fraction = attr.ib()
    female_cloning_fraction = attr.ib()
    male_cloning_fraction = attr.ib()
    sex_ratio = attr.ib()
    bounds_x0 = attr.ib()
    bounds_x1 = attr.ib()
    bounds_y0 = attr.ib()
    bounds_y1 = attr.ib()
    bounds_z0 = attr.ib()
    bounds_z1 = attr.ib()
    migration_records = attr.ib()

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

def decode_population(buff):
    '''
    Extracts the information stored in binary by SLiM in the ``metadata``
    column of a :class:`PopulationTable`. If the buffer is empty, returns None.

    :param bytes buff: The ``metadata`` entry of a row of a
        :class:`PopulationTable`, as stored by SLiM.
    :rtype PopulationMetadata:
    '''
    if len(buff) == 0:
        md = None
    else:
        if len(buff) < 88: # 4 + 8 * 10 + 4
            raise ValueError("Population metadata of incorrect format.")
        num_migration_records = int((len(buff) - 88) / 12) # 4 + 8
        if len(buff) != 88 + 12 * num_migration_records:
            raise ValueError("Population metadata of incorrect format.")
        struct_string = "<i10di" + "id" * num_migration_records
        metadata = struct.unpack(struct_string, buff)
        if metadata[11] != num_migration_records:
            raise ValueError("Inconsistent population metadata format.")
        slim_id = metadata[0]
        selfing_fraction = metadata[1]
        female_cloning_fraction = metadata[2]
        male_cloning_fraction = metadata[3]
        sex_ratio = metadata[4]
        bounds_x0 = metadata[5]
        bounds_x1 = metadata[6]
        bounds_y0 = metadata[7]
        bounds_y1 = metadata[8]
        bounds_z0 = metadata[9]
        bounds_z1 = metadata[10]
        # num_migration_records = metadata[11]
        migration_records = []
        k = 12
        for j in range(num_migration_records):
            source_subpop = metadata[k]
            k += 1
            migration_rate = metadata[k]
            k += 1
            migration_records.append(PopulationMigrationMetadata(source_subpop=source_subpop,
                                                                 migration_rate=migration_rate))

        md = PopulationMetadata(slim_id=slim_id, selfing_fraction=selfing_fraction,
                                  female_cloning_fraction=female_cloning_fraction,
                                  male_cloning_fraction=male_cloning_fraction,
                                  sex_ratio=sex_ratio, bounds_x0=bounds_x0, bounds_x1=bounds_x1,
                                  bounds_y0=bounds_y0, bounds_y1=bounds_y1, bounds_z0=bounds_z0,
                                  bounds_z1=bounds_z1, migration_records=migration_records)
    return md

def encode_population(metadata_object):
    '''
    Encodes :class:`PopulationMetadata` objects as a bytes object, suitable to be put
    in as metadata for a population.  If ``buff`` is ``None``, returns an empty
    bytes object.

    :param PopulationMetadata metadata_object: The object to be encoded.
    :rtype bytes:
    '''
    if metadata_object is None:
        md = b''
    else:
        num_migration_records = len(metadata_object.migration_records)
        for mr in metadata_object.migration_records:
            if type(mr) is not PopulationMigrationMetadata:
                raise ValueError("Migration records should PopulationMigrationMetadata objects.")
        mr_values = [a for b in metadata_object.migration_records for a in b]
        struct_string = "<i10di" + "id" * num_migration_records
        md = struct.pack(struct_string, metadata_object.slim_id,
                         metadata_object.selfing_fraction, metadata_object.female_cloning_fraction,
                         metadata_object.male_cloning_fraction, metadata_object.sex_ratio,
                         metadata_object.bounds_x0, metadata_object.bounds_x1,
                         metadata_object.bounds_y0, metadata_object.bounds_y1,
                         metadata_object.bounds_z0, metadata_object.bounds_z1,
                         num_migration_records, *mr_values)
    return md


def extract_population_metadata(tables):
    '''
    Returns an iterator over lists of :class:`PopulationMetadata` objects
    containing information about the populations in the tables.

    :param TableCollection tables: The tables, as produced by SLiM.
    '''
    metadata = msprime.unpack_bytes(tables.populations.metadata,
                                    tables.populations.metadata_offset)
    for md in metadata:
        yield decode_population(md)


def annotate_population_metadata(tables, metadata):
    '''
    Modify the PopulationTable so that the metadata
    column is given by applying `encode_population()` to the sources given.
    This entirely removes existing information in the Population table.

    :param TableCollection tables: a table collection to be modified
    :param iterable metadata: a list of objects, each PopulationMetadata or None
    '''
    if len(metadata) != tables.populations.num_rows:
        raise ValueError("annotate populations: metadata not the same length as the table.")
    tables.populations.clear()
    for md in metadata:
        if md is None:
            metadata = b''
        else:
            metadata = encode_population(md)
        tables.populations.add_row(metadata=metadata)



"""
Test cases for the metadata reading/writing of pyslim.
"""
from __future__ import print_function
from __future__ import division

import pyslim
import msprime
import tests
import unittest
import os
import tempfile

def get_slim_example_files():
    for filename in ["tests/examples/recipe_WF.trees",
                     "tests/examples/recipe_nonWF.trees"]:
        yield filename

def get_slim_examples():
    for filename in get_slim_example_files():
        yield pyslim.load(filename, slim_format=True)


class TestEncodeDecode(tests.PyslimTestCase):
    '''
    Tests for conversion to/from binary representations of metadata.
    '''

    def test_mutation_metadata(self):
        for md_length in [0, 1, 5]:
            md = [pyslim.MutationMetadata(
                     mutation_type=j, selection_coeff=0.5, population=j,
                     slim_time=10 + j) for j in range(md_length)]
            md_bytes = pyslim.encode_mutation(md)
            new_md = pyslim.decode_mutation(md_bytes)
            self.assertEqual(len(md), len(new_md))
            for x, y in zip(md, new_md):
                self.assertEqual(x, y)

    def test_node_metadata(self):
        md = pyslim.NodeMetadata(slim_id=2, is_null=False,
                                 genome_type=pyslim.GENOME_TYPE_X)
        md_bytes = pyslim.encode_node(md)
        new_md = pyslim.decode_node(md_bytes)
        self.assertEqual(md, new_md)

    def test_individual_metadata(self):
        md = pyslim.IndividualMetadata(
                age=2, pedigree_id=23, population=0,
                sex=pyslim.INDIVIDUAL_TYPE_MALE,
                flags=pyslim.INDIVIDUAL_FLAG_MIGRATED)
        md_bytes = pyslim.encode_individual(md)
        new_md = pyslim.decode_individual(md_bytes)
        self.assertEqual(md, new_md)

    def test_population_metadata(self):
        mrs = [pyslim.PopulationMigrationMetadata(source_subpop=j, migration_rate=0.2)
               for j in range(3)]
        for mr_list in [[], mrs]:
            md = pyslim.PopulationMetadata(
                    slim_id=1, selfing_fraction=0.75, female_cloning_fraction=0.2,
                    male_cloning_fraction=0.8, sex_ratio=0.6, bounds_x0=-1.0,
                    bounds_x1=2.0, bounds_y0=0.0, bounds_y1=0.0, bounds_z0=0.0,
                    bounds_z1=1.0, migration_records=mr_list)
            md_bytes = pyslim.encode_population(md)
            new_md = pyslim.decode_population(md_bytes)
            self.assertEqual(md, new_md)


class TestAnnotate(unittest.TestCase):
    '''
    Tests the table annotation methods.
    '''

    def test_annotate_mutations(self):
        for ts in get_slim_examples():
            tables = ts.tables
            new_tables = ts.tables
            metadata = []
            for md in msprime.unpack_bytes(tables.mutations.metadata, 
                                           tables.mutations.metadata_offset):
                dm = pyslim.decode_mutation(md)
                edm = pyslim.encode_mutation(dm)
                self.assertEqual(md, edm)
                metadata.append(dm)

            pyslim.annotate_mutation_metadata(new_tables, metadata)
            self.assertEqual(tables, new_tables)

    def test_annotate_nodes(self):
        for ts in get_slim_examples():
            tables = ts.tables
            new_tables = ts.tables
            metadata = []
            for md in msprime.unpack_bytes(tables.nodes.metadata, 
                                           tables.nodes.metadata_offset):
                dm = pyslim.decode_node(md)
                edm = pyslim.encode_node(dm)
                self.assertEqual(md, edm)
                metadata.append(dm)

            pyslim.annotate_node_metadata(new_tables, metadata)
            self.assertEqual(tables, new_tables)

    def test_annotate_individuals(self):
        for ts in get_slim_examples():
            tables = ts.tables
            new_tables = ts.tables
            metadata = []
            for md in msprime.unpack_bytes(tables.individuals.metadata, 
                                           tables.individuals.metadata_offset):
                dm = pyslim.decode_individual(md)
                edm = pyslim.encode_individual(dm)
                self.assertEqual(md, edm)
                metadata.append(dm)

            pyslim.annotate_individual_metadata(new_tables, metadata)
            self.assertEqual(tables, new_tables)

    def test_annotate_populations(self):
        for ts in get_slim_examples():
            tables = ts.tables
            new_tables = ts.tables
            metadata = []
            for md in msprime.unpack_bytes(tables.populations.metadata, 
                                           tables.populations.metadata_offset):
                dm = pyslim.decode_population(md)
                edm = pyslim.encode_population(dm)
                self.assertEqual(md, edm)
                metadata.append(dm)

            pyslim.annotate_population_metadata(new_tables, metadata)
            self.assertEqual(tables, new_tables)


class TestDumpLoad(tests.PyslimTestCase):
    '''
    Test reading and writing.
    '''

    def setUp(self):
        fd, self.temp_file = tempfile.mkstemp(prefix="msp_ll_ts_")
        os.close(fd)

    def tearDown(self):
        os.unlink(self.temp_file)

    def verify_times(self, ts, slim_ts):
        gen = slim_ts.slim_generation
        self.assertEqual(ts.num_nodes, slim_ts.num_nodes)
        # verify internal consistency
        for j in range(slim_ts.num_nodes):
            self.assertEqual(slim_ts.node(j).time, 
                             slim_ts.tables.nodes.time[j])
        # verify consistency between tree sequences
        for n1, n2 in zip(ts.nodes(), slim_ts.nodes()):
            self.assertEqual(n1.time, n2.time)

    def verify_dump_equality(self, ts):
        """
        Verifies that we can dump a copy of the specified tree sequence
        to the specified file, and load an identical copy.
        """
        ts.dump(self.temp_file)
        ts2 = pyslim.load(self.temp_file, slim_format=True)
        self.assertEqual(ts.num_samples, ts2.num_samples)
        self.assertEqual(ts.sequence_length, ts2.sequence_length)
        self.assertTablesAlmostEqual(ts.tables, ts2.tables)

    def test_load_tables(self):
        for ts in get_slim_examples():
            self.assertTrue(type(ts) is pyslim.SlimTreeSequence)
            tables = ts.tables
            new_ts = pyslim.load_tables(tables, slim_format=True)
            self.assertTrue(type(new_ts) is pyslim.SlimTreeSequence)
            new_tables = new_ts.tables
            self.assertTablesAlmostEqual(tables, new_tables)

    def test_load(self):
        for fn in get_slim_example_files():
            # unshifted (non-slim) tables
            msp_ts = pyslim.load(fn, slim_format=False)
            self.assertTrue(type(msp_ts) is msprime.TreeSequence)
            msp_tables = msp_ts.tables
            new_ts = pyslim.load_tables(msp_tables, slim_format=True)
            self.assertTrue(type(new_ts) is pyslim.SlimTreeSequence)
            self.verify_times(msp_ts, new_ts)
            new_tables = new_ts.tables
            self.assertTablesAlmostEqual(msp_tables, new_tables)
            # shifted (slim) tables
            slim_ts = pyslim.load(fn, slim_format=True)
            self.assertTrue(type(slim_ts) is pyslim.SlimTreeSequence)
            slim_tables = slim_ts.tables
            new_msp_ts = pyslim.load_tables(slim_tables, slim_format=True)
            self.assertTrue(type(new_msp_ts) is pyslim.SlimTreeSequence)
            self.assertTablesAlmostEqual(slim_tables, new_msp_ts.tables)

    def test_dump_equality(self):
        for ts in get_slim_examples():
            self.verify_dump_equality(ts)



class TestAlleles(tests.PyslimTestCase):
    '''
    Test nothing got messed up with haplotypes.
    '''

    def test_haplotypes(self):
        for slim_ts in get_slim_examples():
            tables = slim_ts.tables
            ts = tables.tree_sequence()
            self.verify_haplotype_equality(ts, slim_ts)




from codecs import open as codecs_open
from setuptools import setup, find_packages
from warnings import warn


# Get the long description from the relevant file
with codecs_open('README.md', encoding='utf-8') as f:
    long_description = f.read()

HAVE_MSPRIME = False
try:
    import msprime
except ImportError:
    warn("`msprime` not present and must be installed")


setup(name='pyslim',
      version='0.0.54',
      description=u"Manipulate tree sequences produced by SLiM.",
      long_description=long_description,
      classifiers=[],
      keywords='',
      author=u"Peter Ralph",
      author_email='petrel.harp@gmail.com',
      url='https://github.com/tskit-dev/pyslim',
      license='MIT',
      packages=find_packages(exclude=[]),
      include_package_data=True,
      zip_safe=False,
      install_requires=['msprime', 'attrs'],
      extras_require={
          'dev': [],
      },
      setup_requires=[],
      )



from __future__ import print_function
from __future__ import division

from pyslim.slim_metadata import *       # NOQA
from pyslim.slim_tree_sequence import *  # NOQA
from pyslim.provenance import *          # NOQA



__version__ = "undefined"
try:
    from . import _version
    __version__ = _version.version
except ImportError:
    pass

def get_provenance_dict():
    """
    Returns a dictionary encoding the version of pyslim used.
    """
    document = {
        "software": "pyslim",
        "version": __version__
    }
    return document


